import java.awt.AlphaComposite;
import java.awt.BorderLayout;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.DisplayMode;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Toolkit;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.WindowEvent;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Random;
import java.util.ResourceBundle;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JPanel;

/*
 ***************************************************************************
 * BOXit 2.0.0                                            2005-02-24 *
 ***************************************************************************
 * Copyright (C) 2001-05 by Frank Hollwitz, hollwitz@users.sourceforge.net *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2 of the License, or (at your  *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it will be useful,         *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with this program; if not, write to the                           *
 * Free Software Foundation, Inc.,                                         *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.               *
 ***************************************************************************
 */

/**
 * @author Frank Hollwitz
 */
public class boxit extends JFrame {

    // declare some variables
    
    // the application preferences
    String strAppTitle = "BOXit";
    String strAppVersion = "2.0.0";
    private Image imgAppIcon = null;
    private Image imgBigIcon = null;
    private Image imgBoss = null;

    // the content panel
    private JPanel contentPane = null;
    // the canvas = the main part of the application
    private boxit_canvas canMain;
    // the timer = the stopwatch of the application
    private volatile boxit_timer timMain = null;

    // graphicsdevice and displaymode
    private GraphicsDevice device;
    private DisplayMode dmdOriginal;
    private DisplayMode dmdGame;
    private boolean isFullscreen = false;
    
    // the position size of the canvas
    private int mainX = 0;
    private int mainY = 0;
    private int mainW = 640;
    private int mainH = 480;
    
    // is a game running?
    private boolean runGame = false;
    // is the running game paused?
    boolean runPause = false;
    // temporary holding pause information
    boolean tmpPause = false;
    
    // the cursors
    private Cursor curDefault;
    private Cursor curHand;
    private Cursor curMove;
    
    // enable sound effects?
    private boolean playSound = true;
    // the sound clips
    private Clip sndSplash = null;
    private Clip sndStart = null;
    private Clip sndClick = null;
    private Clip sndDrag = null;
    private Clip sndDrop = null;
    private Clip sndNodrop = null;
    private Clip sndUndo = null;
    private Clip sndKey = null;
    private Clip sndNohigh = null;
    private Clip sndHigh = null;
    
    // the counter for the available tokens
    private byte countToken = 0;
    // the counters for the elapsed minutes and seconds
    private byte countTimeM = 0;
    private byte countTimeS = 0;
    // the undo string
    // contains the informations for undoing moves
    private String strUndo = "";
    
    // format for a number with leading 0
    DecimalFormat leadZero = new DecimalFormat("00");
    
    // the file for saving options and highscores
    private String strIniFile = ".boxit.ini";
    // the browser for the help file if the standard browser can't detected 
    static String strBrowser = "mozilla";
    
    // the players name
    private String strName = "Player";
    private int lenNameMax = 16;
    // the highscore data
    private ArrayList dataScore = new ArrayList();
    // the first highscore to be shown on the list
    private int shwScore = 1;
    // the as-new-maked highscore to be shown on the list
    private int mrkScore = 0;
    // the scores per page 
    private int pagScore = 11;
    // the highscore length
    private int lenScore = 0;
    // the max. highscore length
    private int maxScore = 100;
    
    // is quitting the game confirmed?
    private boolean confirmQuit = false;
    
    // the text resources
    private ResourceBundle messages = ResourceBundle.getBundle("boxit");
    
    // the image for double buffering
    // All the output is done on this invisible image and then the complete
    // image is copied to the canvas. This is to make sure the canvas is
    // flicker-free.
    private BufferedImage bi;
    // the graphics object of the image
    private Graphics2D big;
    // the transform data of the graphics
    private AffineTransform saveAT;
    
    // the fonts used for the buttons and texts
    private Font fntButton = new Font("Helvetica", Font.BOLD, 15);
    private Font fntText = new Font("Dialog", Font.BOLD, 15);
    private Font fntTitle = new Font("Dialog", Font.BOLD, 18);
    private Font fntCopyr = new Font("Dialog", Font.BOLD, 12);
    // the font colors
    private Color colButtonLight = new Color(192, 192, 192);
    private Color colButtonShadow = new Color(128, 128, 128);
    private Color colButtonDisabled = new Color(160, 160, 160);
    private Color colButtonNormal = new Color(160, 0, 0);
    private Color colButtonOver = new Color(192, 64, 0);
    private Color colButtonClick = new Color(255, 128, 0);
    private Color colText = new Color(0, 128, 64);
    private Color colTextShadow = new Color(192, 255, 192);
    private Color colInput = new Color(32, 64, 128);
    private Color colInputShadow = new Color(192, 192, 255);

    // the mouse position
    private int mouseX;
    private int mouseY;
    
    // which token is pointed by the mouse?
    // 0 = no token
    private int overToken = 0;
    // which token is holded by the mouse?
    // 0 = no token
    private int holdToken = 0;

    // is the mouse over a button?
    // button value + 1
    private int overButton = 0;
    // is the mouse holding a button?
    // button value + 1
    private int holdButton = 0;
    
    // the background image
    private Image imgBackground;
    // the dialog1 (one button) background image
    private Image imgDialog1;
    // the dialog2 (two buttons) background image
    private Image imgDialog2;
    // the highscore dialog background image
    private Image imgHighscore;
    // the about dialog background image
    private Image imgAbout;
    
    // the position and size of the question dialog
    private int questionW = 440;
    private int questionH = 180;
    private int questionX = (mainW - questionW) / 2;
    private int questionY = (mainH - questionH) / 2;

    // the position and size of the highscore dialog
    private int highscoreW = 440;
    private int highscoreH = 440;
    private int highscoreX = (mainW - highscoreW) / 2;
    private int highscoreY = (mainH - highscoreH) / 2;
    
    // the position and size of the question dialog
    private int aboutW = 440;
    private int aboutH = 330;
    private int aboutX = (mainW - aboutW) / 2;
    private int aboutY = (mainH - aboutH) / 2;

    // the token images
    private Image[][] imgToken = new Image[5][5];
    // the tokens rectangles
    // The rectangles are used for detecting if the mouse is over an object.
    private Rectangle[][] rctToken = new Rectangle[5][5];
    private Rectangle rctField;
    // the tokens values
    private int[][] valToken = {{11, 12, 13, 14, 15},
                                {21, 22, 23, 24, 25},
                                {31, 32, 33, 34, 35},
                                {41, 42, 43, 44, 45},
                                {51, 52, 53, 54, 55}};
    
    // the nodrop image
    private Image imgNodrop;
    
    // is a popup visible and which one?
    // find the values below
    private byte showPopup = 0;
    // how many buttons are on every single popup?
    private int[] btnPopup = {9,   // 0 - main menu (this is no popup)
                              2,   // 1 - confirm quit 
                              2,   // 2 - gameover question
                              1,   // 3 - input name
                              8,   // 4 - highscores
                              1,   // 5 - fullscreen error
                              2,   // 6 - confirm cancel game
                              2,   // 7 - confirm highscore reset
                              2,   // 8 - about dialog
                              1    // 9 - timeout dialog
                              };
    // on which index does the buttons for a popup start and end (will be calculated later)?
    private int[] begPopup = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    private int numPopup = btnPopup.length;
    
    // the buttons resources strings
    private String[] resButton = {"ButtonMainNewgame",      //  0 - main - start/stop
                                  "ButtonMainPause",        //  1 - main - pause/resume
                                  "ButtonMainUndo",         //  2 - main - undo
                                  "ButtonMainHighscores",   //  3 - main - highscores
                                  "ButtonMainSound",        //  4 - main - sound
                                  "ButtonMainFullscreen",   //  5 - main - fullscreen/windowed
                                  "ButtonMainHelp",         //  6 - main - help 
                                  "ButtonMainAbout",        //  7 - main - about
                                  "ButtonMainQuit",         //  8 - main - quit
                                  "ButtonQuitNo",           //  9 - confirm quit - no
                                  "ButtonQuitYes",          // 10 - confirm quit - yes
                                  "ButtonGameoverUndo",     // 11 - gameover question - undo
                                  "ButtonGameoverFinish",   // 12 - gameover question - finish
                                  "ButtonNameOk",           // 13 - input name - ok
                                  "ButtonHighscoreReset",   // 14 - highscores - reset
                                  "ButtonHighscoreClose",   // 15 - highscores - close
                                  "ButtonHighscoreFirst",   // 16 - highscores - |<
                                  "ButtonHighscorePgup",    // 17 - highscores - <<
                                  "ButtonHighscorePre",     // 18 - highscores - <
                                  "ButtonHighscoreNext",    // 19 - highscores - >
                                  "ButtonHighscorePgdn",    // 20 - highscores - >>
                                  "ButtonHighscoreLast",    // 21 - highscores - >|
                                  "ButtonFullscreenOk",     // 22 - fullscreen - ok
                                  "ButtonQuitNo",           // 23 - confirm quit - no
                                  "ButtonQuitYes",          // 24 - confirm quit - yes
                                  "ButtonQuitNo",           // 25 - confirm highscore reset - no
                                  "ButtonQuitYes",          // 26 - confirm highscore reset - yes
                                  "ButtonAboutLegal",       // 27 - about dialog - no
                                  "ButtonAboutClose",       // 28 - about dialog - yes
                                  "ButtonGameoverFinish"    // 29 - timeout dialog - finish
                                  };
    // the buttons active states
    // 0 = disabled
    // 1 = normal
    // 2 = pointed by the mouse
    // 3 = clicked
    private int[] actButton = {1,   //  0
                               0,   //  1 
                               0,   //  2
                               1,   //  3
                               1,   //  4
                               1,   //  5 
                               1,   //  6
                               1,   //  7
                               1,   //  8
                               1,   //  9
                               1,   // 10
                               1,   // 11
                               1,   // 12
                               1,   // 13
                               1,   // 14
                               1,   // 15
                               1,   // 16
                               1,   // 17
                               1,   // 18
                               1,   // 19
                               1,   // 20
                               1,   // 21
                               1,   // 22
                               1,   // 23
                               1,   // 24
                               1,   // 25
                               1,   // 26
                               1,   // 27
                               1,   // 28
                               1    // 29
                               }; 
    // the button rectangles
    private Rectangle[] rctButton = new Rectangle[actButton.length];
    
    // the digit images
    private Image[] imgDigit = new Image[12];

    // the coordinates of the tokens
    private int[] tokenX = {86, 165, 244, 323, 402};
    private int[] tokenY = {46, 125, 204, 283, 362};
    private int tokenW = 72;
    private int tokenH = 72;
    private int tokenR = 3;

    // the coordinates of the buttons
    private int[] buttonX = {500,               //  0
                             500,               //  1
                             500,               //  2
                             500,               //  3
                             500,               //  4
                             500,               //  5
                             500,               //  6
                             500,               //  7
                             500,               //  8
                             questionX + 30,    //  9
                             questionX + 290,   // 10
                             questionX + 30,    // 11
                             questionX + 290,   // 12
                             questionX + 290,   // 13
                             highscoreX + 30,   // 14
                             highscoreX + 290,  // 15
                             highscoreX + 40,   // 16
                             highscoreX + 80,   // 17
                             highscoreX + 120,  // 18
                             highscoreX + 290,  // 19
                             highscoreX + 330,  // 20
                             highscoreX + 370,  // 21
                             questionX + 290,   // 22
                             questionX + 30,    // 23
                             questionX + 290,   // 24
                             questionX + 30,    // 25
                             questionX + 290,   // 26
                             aboutX + 30,       // 27
                             aboutX + 290,      // 28
                             questionX + 290    // 29
                             };
    private int[] buttonY = {40,                //  0
                             85,                //  1
                             130,               //  2
                             175,               //  3
                             225,               //  4
                             270,               //  5
                             320,               //  6
                             365,               //  7
                             410,               //  8
                             questionY + 120,   //  9
                             questionY + 120,   // 10
                             questionY + 120,   // 11
                             questionY + 120,   // 12
                             questionY + 120,   // 13
                             highscoreY + 380,  // 14
                             highscoreY + 380,  // 15
                             highscoreY + 320,  // 16
                             highscoreY + 320,  // 17
                             highscoreY + 320,  // 18
                             highscoreY + 320,  // 19
                             highscoreY + 320,  // 20
                             highscoreY + 320,  // 21
                             questionY + 120,   // 22
                             questionY + 120,   // 23
                             questionY + 120,   // 24
                             questionY + 120,   // 25
                             questionY + 120,   // 26
                             aboutY + 270,      // 27
                             aboutY + 270,      // 28
                             questionY + 120    // 29
                             };
    private int buttonW = 120;
    private int buttonWs = 30;
    private int buttonH = 30;
    
    // the coordinates of the text outputs
    private int textDialogX = questionX + 30;
    private int textDialogY = questionY + 30;
    private int textDialogW = 380;
    private int textDialogH = 70;
    private int textHighNumX = highscoreX + 30;
    private int textHighNumW = 40;
    private int textHighNameX = highscoreX + textHighNumW + 30;
    private int textHighNameW = 190;
    private int textHighTokensX = highscoreX + textHighNumW + textHighNameW + 30;
    private int textHighTokensW = 75;
    private int textHighTimeX = highscoreX + textHighNumW + textHighNameW + textHighTokensW + 30;
    private int textHighTimeW = 75;
    private int textHighHeadY = highscoreY + 30;
    private int textHighHeadH = 30;
    private int textHighLineY = highscoreY + 65;
    private int textHighLineH = 30;
    private int textHighFootX = highscoreX + 160;
    private int textHighFootW = 120;
    private int textHighFootY = highscoreY + 320;
    private int textHighFootH = 30;
    
    // the size of the nodrop image
    private int nodropW = 32;
    private int nodropH = 32;
    
    // the position and size of the status line
    private int textStatusX = 35;
    private int textStatusY = 55;
    private int textStatusW = 40;
    private int textStatusH = 370;

    // the position of the counters
    private int countTokenX = 39;
    private int countTokenY = 140;
    private int countTimeX = 56;
    private int countTimeY = 140;
    private int countDigitH = 12;
    
    // is a making move valid?
    private boolean moveValid = false;

    private KeyListener keyMain = null;
    
    public boxit () {

        keyMain = new KeyListener() { 
            public void keyPressed(KeyEvent e) {
                // bosskey - B
                if ((e.getKeyCode() == KeyEvent.VK_B) && (e.getModifiers() == 0)) {
                    BossIsComing();
                }
                switch (showPopup) {
                case 0:
                    // no popup dialog
                    // start a new game - F2, N
                    if (((e.getKeyCode() == KeyEvent.VK_F2) || (e.getKeyCode() == KeyEvent.VK_N)) && (e.getModifiers() == 0)) {
                        if (!runGame) {
                            clickButtonNew();
                        }
                    }
                    // cancel a running game - F2, A
                    if ((e.getKeyCode() == KeyEvent.VK_A) && (e.getModifiers() == 0)) {
                        if (runGame) {
                            clickButtonNew();
                        }
                    }
                    // pause a game - F3, PAUSE, P
                    if (((e.getKeyCode() == KeyEvent.VK_F3) || (e.getKeyCode() == KeyEvent.VK_PAUSE) || (e.getKeyCode() == KeyEvent.VK_P) || (e.getKeyCode() == KeyEvent.VK_SPACE)) && (e.getModifiers() == 0)) {
                        if (actButton[1] != 0) {
                            doSound(sndClick);
                            clickButtonPause();
                        }
                    }
                    // undo a move - F8, BACKSPACE, DEL
                    if (((e.getKeyCode() == KeyEvent.VK_F8) || (e.getKeyCode() == KeyEvent.VK_BACK_SPACE) || (e.getKeyCode() == KeyEvent.VK_DELETE)) && (e.getModifiers() == 0)) {
                        if (actButton[2] != 0) {
                            clickButtonUndo();
                        }
                    }
                    // show highscores - F6
                    if ((e.getKeyCode() == KeyEvent.VK_F6) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        clickButtonScore();
                    }
                    // change sound - F7, S
                    if (((e.getKeyCode() == KeyEvent.VK_F7) || (e.getKeyCode() == KeyEvent.VK_S)) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        clickButtonSound();
                    }
                    // change fullscreen - F4
                    if ((e.getKeyCode() == KeyEvent.VK_F4) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        clickButtonFullscreen();
                    }
                    // show help - F1, H
                    if (((e.getKeyCode() == KeyEvent.VK_F1) || (e.getKeyCode() == KeyEvent.VK_H)) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        clickButtonHelp();
                    }
                    // show about - F9
                    if ((e.getKeyCode() == KeyEvent.VK_F9) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        clickButtonInfo();
                    }
                    // quit - F10
                    if ((e.getKeyCode() == KeyEvent.VK_F10) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        clickButtonQuit();
                    }
                    break;
                case 1:
                    // confirm quit dialog
                    // no - ESC
                    if ((e.getKeyCode() == KeyEvent.VK_ESCAPE) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        showPopup = 0;
                    }
                    // yes - RET
                    if ((e.getKeyCode() == KeyEvent.VK_ENTER) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        confirmQuit = true;
                        clickButtonQuit();
                    }
                    break;
                case 2:
                    // confirm gameover dialog
                    // undo - ESC
                    if ((e.getKeyCode() == KeyEvent.VK_ESCAPE) && (e.getModifiers() == 0)) {
                        showPopup = 0;
                        doPause();
                        clickButtonUndo();
                    }
                    // finish - RET
                    if ((e.getKeyCode() == KeyEvent.VK_ENTER) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        gameOver();
                    }
                    break;
                case 3:
                    // name input dialog
                    // ok - RET
                    if ((e.getKeyCode() == KeyEvent.VK_ENTER) && (e.getModifiers() == 0)) {
                        if (actButton[13] != 0) {
                            doSound(sndClick);
                            showPopup = 4;
                        }
                    }
                    break;
                case 4:
                    // highscore dialog
                    int p = 0;
                    if (lenScore > 0) {
                        p = lenScore - pagScore + 1;
                        if (p > lenScore) { p = lenScore; } 
                        if (p < 1) { p = 1; } 
                    }
                    // reset - DEL
                    if ((e.getKeyCode() == KeyEvent.VK_DELETE) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        showPopup = 7;
                    }
                    // close - RET
                    if ((e.getKeyCode() == KeyEvent.VK_ENTER) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        mrkScore = 0;
                        showPopup = 0;
                    }
                    // |< - HOME
                    if ((e.getKeyCode() == KeyEvent.VK_HOME) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        if (lenScore > 0) {
                            shwScore = 1;
                        }
                    }
                    // << - PGUP
                    if ((e.getKeyCode() == KeyEvent.VK_PAGE_UP) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        if (lenScore > 0) {
                            shwScore -= 10;
                            if (shwScore < 1) { shwScore = 1; }
                        }
                    }
                    // < - UP, LEFT
                    if (((e.getKeyCode() == KeyEvent.VK_UP) || (e.getKeyCode() == KeyEvent.VK_LEFT)) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        if (lenScore > 0) {
                            shwScore--;
                            if (shwScore < 1) { shwScore = 1; }
                        }
                    }
                    // > - DOWN, RIGHT
                    if (((e.getKeyCode() == KeyEvent.VK_DOWN) || (e.getKeyCode() == KeyEvent.VK_RIGHT)) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        if (lenScore > 0) {
                            shwScore++;
                            if (shwScore > p) { shwScore = p; }
                        }
                    }
                    // >> - PGDOWN
                    if ((e.getKeyCode() == KeyEvent.VK_PAGE_DOWN) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        if (lenScore > 0) {
                            shwScore += 10;
                            if (shwScore > p) { shwScore = p; }
                        }
                    }
                    // >| - END
                    if ((e.getKeyCode() == KeyEvent.VK_END) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        if (lenScore > 0) {
                            shwScore = p;
                        }
                    }
                    break;
                case 5:
                    // fullscreen error
                    // ok - RET
                    if ((e.getKeyCode() == KeyEvent.VK_ENTER) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        showPopup = 0;
                    }
                    break;
                case 6:
                    // confirm cancel game dialog
                    // no - ESC
                    if (((e.getKeyCode() == KeyEvent.VK_ESCAPE) || (e.getKeyCode() == KeyEvent.VK_F8) || (e.getKeyCode() == KeyEvent.VK_BACK_SPACE) || (e.getKeyCode() == KeyEvent.VK_DELETE)) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        if (!tmpPause) { doPause(); }
                        showPopup = 0;
                    }
                    // yes - RET
                    if ((e.getKeyCode() == KeyEvent.VK_ENTER) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        showPopup = 0;
                        gameOver();
                    }
                    break;
                case 7:
                    // confirm highscore reset
                    // no - ESC
                    if ((e.getKeyCode() == KeyEvent.VK_ESCAPE) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        showPopup = 4;
                    }
                    // yes - RET
                    if ((e.getKeyCode() == KeyEvent.VK_ENTER) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        if (lenScore > 0) {
                            mrkScore = 0;
                            lenScore = 0;
                            dataScore.clear();
                        }
                        showPopup = 4;
                    }
                    break;
                case 8:
                    // about dialog
                    // license - L
                    if ((e.getKeyCode() == KeyEvent.VK_L) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        try { BrowserLauncher.openURL(getFileURL("license")); } catch (Exception x) {}
                    }
                    // close - RET
                    if ((e.getKeyCode() == KeyEvent.VK_ENTER) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        showPopup = 0;
                    }
                    break;
                case 9:
                    // timeout dialog
                    // finish - RET
                    if ((e.getKeyCode() == KeyEvent.VK_ENTER) && (e.getModifiers() == 0)) {
                        doSound(sndClick);
                        showPopup = 0;
                        gameOver();
                    }
                    break;
                }
                canMain.repaint();
            }
            public void keyTyped(KeyEvent e) {
                if (showPopup == 3) {
                    int l = strName.length();
                    char c = e.getKeyChar();
                    switch (c) {
                    case 8:
                    case 127: 
                        if (l > 0) {
                            strName = strName.substring(0, l - 1);
                            doSound(sndKey);
                        }
                        break;
                    case 10:
                    case 13: 
                        break;
                    default:
                        if (l < lenNameMax) {
                            strName = strName + c;
                            doSound(sndKey);
                        }
                        break;
                    }
                    if ((strName.length() == 0) || (strName.length() > lenNameMax)) { actButton[13] = 0; } else { actButton[13] = 1; }
                    canMain.repaint();
                }
            }
            public void keyReleased(KeyEvent e) {}
        };
        addKeyListener(keyMain);
        
        // getting graphics devices
        GraphicsEnvironment env = GraphicsEnvironment.getLocalGraphicsEnvironment();
        device = env.getDefaultScreenDevice();
        dmdOriginal = device.getDisplayMode();
        dmdGame = new DisplayMode(mainW, mainH, 32, DisplayMode.REFRESH_RATE_UNKNOWN);
        
        // set the filename for the inifile
        String dirUser = System.getProperty("user.home", "");
        if (dirUser.length() != 0) { strIniFile = dirUser + File.separator + strIniFile; }
        
        // getting the options and highscores from the inifile
        JIniFile iniFile = new JIniFile(strIniFile);
        String sn = iniFile.ReadString("Player", "Name", strName);
        if (sn != "") { strName = sn; }
        strBrowser = iniFile.ReadString("Help", "Browser", strBrowser);
        mainX = iniFile.ReadInteger("Form", "Left", 0);
        mainY = iniFile.ReadInteger("Form", "Top", 0);
        int full = iniFile.ReadInteger("Form", "Fullscreen", 0);
        int ps = iniFile.ReadInteger("Options", "Sound", 1);
        if (ps == 0) { playSound = false; } else { playSound = true; }
        lenScore = iniFile.ReadInteger("Options", "Scores", 0);
        if (lenScore > 0) {
            String[] setScore = new String[3];
            for (int i = 0; i < lenScore; i++) {
                setScore[0] = iniFile.ReadString("Scores", Integer.toString(i + 1) + "-Name", "");
                setScore[1] = iniFile.ReadString("Scores", Integer.toString(i + 1) + "-Tokens", "");
                setScore[2] = iniFile.ReadString("Scores", Integer.toString(i + 1) + "-Time", "");
                if ((setScore[0].length() > 0) && (setScore[1].length() > 0) && (setScore[2].length() > 0)) {
                    dataScore.add(setScore.clone());
                }
            }
            lenScore = dataScore.size();
        }
        iniFile = null;

        // on which index does the buttons for a popup start and end?
        for (int i = 1; i < numPopup; i++) { begPopup[i] = begPopup[i - 1] + btnPopup[i - 1]; }

        // getting the application icon
        imgAppIcon = loadImage("icon").getImage();
        imgBigIcon = loadImage("boxit").getImage();
        imgBoss = loadImage("boss").getImage();
        
        // setting the cursors
        curDefault = new Cursor(Cursor.DEFAULT_CURSOR);
        curHand = new Cursor(Cursor.HAND_CURSOR);
        curMove = Toolkit.getDefaultToolkit().createCustomCursor(loadImage("cursornone").getImage(), new Point(0, 0), "cursornone");
        
        // getting the sounds
        sndSplash = loadSound("splash");
        sndStart = loadSound("start");
        sndClick = loadSound("click");
        sndDrag = loadSound("drag");
        sndDrop = loadSound("drop");
        sndNodrop = loadSound("nodrop");
        sndUndo = loadSound("undo");
        sndKey = loadSound("key");
        sndNohigh = loadSound("nohigh");
        sndHigh = loadSound("high");

        //setting the window location
        this.setLocation(mainX, mainY);
        // setting the window title
        this.setTitle(strAppTitle + " " + strAppVersion);
        // setting the applications icon
        this.setIconImage(imgAppIcon);
        
        // setting the contentpane
        contentPane = (JPanel) this.getContentPane();
        contentPane.setPreferredSize(new Dimension(mainW, mainH));
        contentPane.setLayout(new BorderLayout());

        // setting the canvas
        canMain = new boxit_canvas();
        contentPane.add(canMain, BorderLayout.CENTER);
        
        // getting the background image
        imgBackground = loadJpeg("back").getImage();
        // getting the dialog1 background image
        imgDialog1 = loadJpeg("dialog1").getImage();
        // getting the dialog2 background image
        imgDialog2 = loadJpeg("dialog2").getImage();
        // getting the highscore dialog background image
        imgHighscore = loadJpeg("high").getImage();
        // getting the about dialog background image
        imgAbout = loadJpeg("about").getImage();
        
        // getting the nodrop image
        imgNodrop = loadImage("nodrop").getImage();
        
        // getting the digit images
        for (int i = 0; i < 12; i++) { imgDigit[i] = loadImage("digit" + Integer.toString(i)).getImage(); }
        
        // getting the token images
        for (int i = 0; i < 5; i++) {
            for (int j = 0; j < 5; j++) {
                imgToken[i][j] = loadImage("token" + Integer.toString(i + 1) + Integer.toString(j + 1)).getImage();
            }
        }

        // make the window unable to resized by the user
        this.setResizable(false);
        // setting the size of the frame
        this.pack();
        
        // showing the window
        this.setVisible(true);
        
        // make this fullscreen if the options says it
        if (full == 1) { doFullscreen(); }
        
        doSound(sndSplash);

    }

    // the void main = where everything starts 
    public static void main(String[] args) {
        
        // starting the application by creating the window
        boxit fraMain = new boxit();

    }
    
    // enter fullscreen mode
    private void doFullscreen () {
        
        if ((runGame) && (!runPause)) { doPause(); }
        
        boolean b = isFullscreen;
        if (b) {
            doWindow();
        }
        else {
            canMain.setVisible(false);
            mainX = this.getX();
            mainY = this.getY();
            boolean error = false;
            this.dispose();
            this.setUndecorated(true);
            isFullscreen = true;
            if (device.isFullScreenSupported()) {
                try {
                    device.setFullScreenWindow(this);
                } catch (Exception e) { error = true; }
                if (device.isDisplayChangeSupported()) {
                    try {
                        device.setDisplayMode(dmdGame);
                    } catch (Exception e) { error = true; }
                }
                else { error = true; }
            }
            else { error = true; }
            if (error) {
                doWindow();
                showPopup = 5;
            }
            canMain.setVisible(true);
        }
    }
    
    // leave fullscreen mode
    private void doWindow () {
        
        if (isFullscreen) {
            try {
                device.setDisplayMode(dmdOriginal);
            } catch (Exception e) {}
            try {
                device.setFullScreenWindow(null);
            } catch (Exception e) {}
            this.dispose();
            this.setUndecorated(false);
            this.setLocation(mainX, mainY);
            this.pack();
            this.setVisible(true);
            isFullscreen = false;
        }
        canMain.repaint();
        
    }
    
    // outputs the webbrowser to use if no standard browser is detectable
    public static String getBrowser() {
        return strBrowser;
    }

    // handling window events
    protected void processWindowEvent(WindowEvent e) {
        
        // closing the window = closing the application
        if (e.getID() == WindowEvent.WINDOW_CLOSING) {

            doSound(sndClick);
            clickButtonQuit();

        } else
        // activating the window = the window gets the focus
        if (e.getID() == WindowEvent.WINDOW_ACTIVATED) {

            //

        } else
        // deactivating the window = the windows loses the focus
        if (e.getID() == WindowEvent.WINDOW_DEACTIVATED) {

            if (!runPause) { doPause(); }

        } else
        // deiconifiing the window = the window is rising from the taskbar
        if (e.getID() == WindowEvent.WINDOW_DEICONIFIED) {

            BossIsGone();

        }
        
    }

    // the boss key (escape) is pressed
    private void BossIsComing() {

        this.setVisible(false);
        this.setIconImage(imgBoss);
        this.setTitle("Unbenannt - Editor");
        this.setVisible(true);
        if (this.getState() != Frame.ICONIFIED) { this.setState(Frame.ICONIFIED); }

    }

    // the boss is gone
    private void BossIsGone() {

        this.setVisible(false);
        this.setIconImage(imgAppIcon);
        this.setTitle(strAppTitle + " " + strAppVersion);
        this.setVisible(true);

    }
    
    // loading the images
    private ImageIcon loadImage (String f) {
        
        ImageIcon img = new ImageIcon(getClass().getResource("images/" + f + ".png"));
        return img;
        
    }
    
    // loading the jpeg images
    private ImageIcon loadJpeg (String f) {
        
        ImageIcon img = new ImageIcon(getClass().getResource("images/" + f + ".jpg"));
        return img;
        
    }
    
    // loading the sound clips
    private Clip loadSound (String f) {
        
        Clip clip = null;
        clip = getSound("empty");
        clip.flush();
        clip.close();
        clip = getSound(f);
        return clip;
        
    }
    
    // getting the sound clips
    private Clip getSound (String f) {
        
        Clip clip = null;
        
        try {
            // creating an audio stream with a sound
            AudioInputStream stream = AudioSystem.getAudioInputStream(getClass().getResource("sounds/" + f + ".wav"));

            // make sure the audio format is playable by java
            AudioFormat format = stream.getFormat();
            if ((format.getEncoding() == AudioFormat.Encoding.ULAW) ||
                (format.getEncoding() == AudioFormat.Encoding.ALAW)) {
                AudioFormat newFormat = new AudioFormat(AudioFormat.Encoding.PCM_SIGNED,
                                                             format.getSampleRate(),
                                                             format.getSampleSizeInBits()*2,
                                                             format.getChannels(),
                                                             format.getFrameSize()*2,
                                                             format.getFrameRate(), true);
                stream = AudioSystem.getAudioInputStream(newFormat, stream);
                format = newFormat;
            }

            // creating a data line and give the streams data to the clip
            DataLine.Info info = new DataLine.Info(Clip.class, format);
            clip = (Clip) AudioSystem.getLine(info);
            clip.open(stream);
        }
        // handling the errors
        catch (UnsupportedAudioFileException audioException) {}
        catch (LineUnavailableException noLineException) {}
        catch (IOException ioException) {}
        catch (Exception e) {}
        
        return clip;
        
    }
    
    // playing the sounds
    private void doSound (Clip clip) {
        
        // is the sound enabled? 
        if (playSound) {
            // is a playable clip available?
            if (clip != null) {
                // set the clips starting position...
                clip.setFramePosition(0);
                // ...and play the clip 
                clip.start();
            }
        }
        
    }
    
    // pause the game
    private void doPause () {
        
        boolean b = runPause;
        if (runGame) {
            if (!b) {
                runPause = true;
                
            }
            else {
                
                runPause = false;
            }
        }
        
    }

    // the user clicked the new button
    private void clickButtonNew () {
        
        // starting a new game
        if (!runGame) {

            doSound(sndStart);

            // resetting the tokens
            for (int i = 0; i < 5; i++) {
                for (int j = 0; j < 5; j++) {
                    valToken[i][j] = ((i + 1) * 10) + (j + 1);
                }
            }

            // mixing the tokens
            Random rangen = new Random();
            for (int i = 0; i < 50; i++) {
                int r1 = rangen.nextInt(5);
                int r2 = rangen.nextInt(5);
                int r3 = rangen.nextInt(5);
                int r4 = rangen.nextInt(5);
                int dummy = valToken[r1][r2];
                valToken[r1][r2] = valToken[r3][r4];
                valToken[r3][r4] = dummy;
            }
            
            // reset the counters and switches
            countToken = 25;
            countTimeM = 0;
            countTimeS = 0;
            strUndo = "";
            runPause = false;
            runGame = true;
            
            // enable the pause button
            actButton[1] = 1;
            
            // initializing the timer
            timMain = new boxit_timer();
            timMain.start();
            
        }
        
        // cancel the game
        else {
            
            doSound(sndClick);

            tmpPause = runPause;
            if (!runPause) { doPause(); }
            showPopup = 6;
            
        }
        
        
    }
    
    // the user clicked the pause button
    private void clickButtonPause () {
        
        // (un)pause the game
        doPause();
        
    }
    
    // the user clicked the undo button
    private void clickButtonUndo () {
        
        if ((runGame) && (!runPause) && (strUndo != "")) {
            // getting informations from undo string
            int valoc = Integer.parseInt(strUndo.substring(0, 2));
            int valhc = Integer.parseInt(strUndo.substring(2, 4));
            int valpv = Integer.parseInt(strUndo.substring(4, 6));
            // calculating the move to be undone
            int ocx = getTokenX(valoc);
            int ocy = getTokenY(valoc);
            int hcx = getTokenX(valhc);
            int hcy = getTokenY(valhc);
            // reset the moved token
            valToken[ocx][ocy] = valToken[hcx][hcy];
            // set the removed token
            valToken[hcx][hcy] = valpv;
            // cut the undo string
            if (strUndo.length() > 6) { strUndo = strUndo.substring(6); } else { strUndo = ""; }
            // count down the number of available tokens
            countToken++;
            if (strUndo == "") { actButton[2] = 0; } else { actButton[2] = 1; }
            // setting the marker for move check
            moveValid = false;
            // playing the undo sound
            doSound(sndUndo);
            
        }

    }
    
    // the user clicked the score button
    private void clickButtonScore () {
        
        // is there no other popup?
        if (showPopup == 0) {
            if ((runGame) && (!runPause)) { doPause(); }
            showPopup = 4;
            canMain.repaint();
        }
        
    }
    
    // the user clicked the sound button
    private void clickButtonSound () {
        
        // switching the sound
        playSound = !playSound;
        
    }
    
    // the user clicked the fullscreen button
    private void clickButtonFullscreen () {
        
        doFullscreen();
        
    }
    
    // the user clicked the help button
    private void clickButtonHelp () {
        
        if ((runGame) && (!runPause)) { doPause(); }
        try { BrowserLauncher.openURL(getFileURL("")); } catch (Exception x) {}
        
    }
    
    // the user clicked the about button
    private void clickButtonInfo () {
        
        if ((runGame) && (!runPause)) { doPause(); }
        showPopup = 8;
        
    }
    
    // the user clicked the quit button
    private void clickButtonQuit () {
        
        // closing the application = where everything ends
        if (confirmQuit) {
            
            // writing the options and highscores to the inifile
            JIniFile iniFile = new JIniFile(strIniFile);
            iniFile.WriteString("Player", "Name", strName);
            // iniFile.WriteString("Help", "Browser", strBrowser);
            if (!isFullscreen) {
                mainX = this.getX();
                mainY = this.getY();
            }
            iniFile.WriteInteger("Form", "Left", mainX);
            iniFile.WriteInteger("Form", "Top", mainY);
            int full = 0; if (isFullscreen) { full = 1; }
            iniFile.WriteInteger("Form", "Fullscreen", full);
            int ps = 0; if (playSound) { ps = 1; }
            iniFile.WriteInteger("Options", "Sound", ps);
            iniFile.WriteInteger("Options", "Scores", lenScore);
            iniFile.EraseSection("Scores");
            iniFile.UpdateFile();
            if (lenScore > 0) {
                String[] setScore = new String[3];
                for (int i = 0; i < lenScore; i++) {
                    setScore = (String[])dataScore.get(i);
                    iniFile.WriteString("Scores", Integer.toString(i + 1) + "-Name", setScore[0]);
                    iniFile.WriteString("Scores", Integer.toString(i + 1) + "-Tokens", setScore[1]);
                    iniFile.WriteString("Scores", Integer.toString(i + 1) + "-Time", setScore[2]);
                }
                iniFile.UpdateFile();
            }
            iniFile = null;

            // back to system
            doWindow();
            System.exit(0);
            
        }
        else {
            // is there no popup
            if (showPopup == 0) {
                if ((runGame) && (!runPause)) { doPause(); }
                showPopup = 1;
                canMain.repaint();
            }
        }
        
    }
    
    // a game runs more than one hour = timeout
    private void timeout () {
        
        doPause();
        showPopup = 9;
        canMain.repaint();
        
    }

    // the game is over
    private void gameOver () {
        
        calcScore();
        
        if (mrkScore > 0) {
            int p = 0;
            if (lenScore > 0) {
                p = lenScore - pagScore + 1;
                if (p > lenScore) { p = lenScore; } 
                if (p < 1) { p = 1; } 
                shwScore = mrkScore - 5;
                if (shwScore < 1) { shwScore = 1; }
                if (shwScore > p) { shwScore = p; }
            }
            doSound(sndHigh);
            showPopup = 3;
        }
        else {
            // no highscore reached
            doSound(sndNohigh);
            showPopup = 4;
        }
        
        // reset the game data
        countToken = 25;
        countTimeM = 0;
        countTimeS = 0;
        strUndo = "";
        runPause = false;
        runGame = false;
        
        timMain.interrupt();
        timMain = null;
        
    }
    
    // the game is over
    private void calcScore () {
        
        lenScore = dataScore.size();

        // dummy arrays to access and to change the score data
        String[] setScore = new String[6];
        String[] getScore = new String[6];
        
        // create a dataset from the new score
        setScore[0] = strName;
        setScore[1] = leadZero.format(countToken);
        setScore[2] = leadZero.format(countTimeM) + ":" + leadZero.format(countTimeS);
        
        if (lenScore == 0) {
            mrkScore = 1;
        }
        else { 
            // sort the highscores
            mrkScore = lenScore + 1;
            for (int i = (lenScore - 1); i >= 0; i--) {
              getScore = (String[])dataScore.get(i);
              int getToken = Integer.parseInt(getScore[1]);
              int countTime = (countTimeM * 60) + countTimeS; 
              int getTime = (Integer.parseInt(getScore[2].substring(0, 2)) * 60) + Integer.parseInt(getScore[2].substring(3, 5));
              if ((countToken < getToken) || ((countToken == getToken) && (countTime < getTime))) {
                  mrkScore = i + 1;
              }
            }
        }

        // adding the new score to highscores if it's not out of range
        if (mrkScore <= maxScore) { dataScore.add(mrkScore - 1, setScore.clone()); } else { mrkScore = 0; }
        lenScore = dataScore.size();
        
    }
    
    // calculating the x coordinate of a token
    private int getTokenX (int i) {
        return (i / 10) - 1;
    }
    
    // calculating the y coordinate of a token
    private int getTokenY (int i) {
        return (i % 10) - 1;
    }
    
    // calculating the url of the needed help file
    private String getFileURL (String n) {
        
        String s = "help";
        String l = "_" + System.getProperty("user.language");
        if ((new File(s + l + ".html")).exists()) { s += l; }
        File f = new File(s + ".html");
        URL u = null;
        try { u = f.toURL(); } catch (Exception x) {}
        s = u.toString();
        if (n != "") { s += "#" + n; }
        return s;

    }
    // the canvas = where all the visible work is done
    private class boxit_canvas extends Canvas implements MouseListener, MouseMotionListener {
        
        public boxit_canvas () {
            
            // setting the event handlers for the canvas' mouse events
            addMouseMotionListener(this);
            addMouseListener(this);
            addKeyListener(keyMain);

            // setting the token rectangles
            for (int i = 0; i < 5; i++) {
                for (int j = 0; j < 5; j++) {
                    rctToken[i][j] = new Rectangle(tokenX[i] - tokenR, tokenY[j] - tokenR, tokenW + (tokenR * 2), tokenH + (tokenR * 2)); 
                }
            }
            rctField = new Rectangle(tokenX[0] - tokenR, tokenY[0] - tokenR, (tokenW + (tokenR * 2)) * 5, (tokenH + (tokenR * 2)) * 5); 

            // setting the button rectangles
            for (int i = 0; i < rctButton.length; i++) {
                rctButton[i] = new Rectangle(buttonX[i], buttonY[i], buttonW, buttonH); 
            }
            for (int i = 16; i < 22; i++) {
                rctButton[i] = new Rectangle(buttonX[i], buttonY[i], buttonWs, buttonH); 
            }
            
            // creating the double buffering image and its graphics object
            bi = new BufferedImage(mainW, mainH, BufferedImage.TYPE_INT_RGB);
            big = bi.createGraphics();
            
            // getting the standard transform information
            saveAT = big.getTransform();
            
        }
        
        // a mouse button is pressed down
        public void mousePressed (MouseEvent e) {
            
            if (e.getButton() == MouseEvent.BUTTON1) {
                
                // getting the mouse coordinates
                mouseX = e.getX();
                mouseY = e.getY();
                
                // is no popup visible?
                if (showPopup == 0) {

                    // is a game running and is this game not paused and is no token
                    // holded by the mouse?
                    if ((runGame) && (!runPause) && (holdToken == 0)) {
                        // checking all tokens
                        for (int i = 0; i < 5; i++) {
                            for (int j = 0; j < 5; j++) {
                                // is the mouse over a token?
                                if (rctToken[i][j].contains(e.getX(), e.getY())) {
                                    if (valToken[i][j] != 0) {
                                        // playing the drag sound
                                        doSound(sndDrag);
                                        // holding the token
                                        holdToken = ((i + 1) * 10) + (j + 1);
                                        // at this time no token is overed
                                        overToken = 0;
                                        // setting the marker for move check
                                        moveValid = true;
                                        // updating the cursor
                                        updateCursor(e);
                                        // updating the canvas
                                        updateLocation(e);
                                    }
                                }
                            }
                        }
                    }
                    
                }
                
                // checking all buttons of the visible popup
                for (int i = begPopup[showPopup]; i < (begPopup[showPopup] + btnPopup[showPopup]); i++) {
                    if (rctButton[i].contains(e.getX(), e.getY())) {
                        // is this button enabled?
                        if (actButton[i] > 0) {
                            // holding the button
                            holdButton = i + 1;
                            actButton[i] = 3; 
                            // updating the canvas
                            updateLocation(e);
                        }
                    }
                }
                
            }
            
            if (e.getButton() == MouseEvent.BUTTON3) {
                
                if ((runGame) && (!runPause) && (showPopup == 0)) {
                    // was the click within the playing field?
                    if (rctField.contains(e.getX(), e.getY())) {
                        // is the undo button enabled?
                        if (actButton[2] == 1) {
                            // holding the undo button
                            holdButton = 101;
                            actButton[2] = 3;
                            // updating the canvas
                            updateLocation(e);
                        }
                    }
                }
                
            }
            
        }
        
        // the mouse is dragged while a mouse button is pressed down
        public void mouseDragged (MouseEvent e) {
            
            // getting the mouse coordinates
            mouseX = e.getX();
            mouseY = e.getY();
            
            // is no popup visible?
            if (showPopup == 0) {

                // is a game running and is this game not paused and is a token
                // holded by the mouse?
                if ((runGame) && (!runPause) && (holdToken > 0)) {
                    // checking all tokens
                    overToken = 0;
                    for (int i = 0; i < 5; i++) {
                        for (int j = 0; j < 5; j++) {
                            // is the mouse over a token?
                            if (rctToken[i][j].contains(e.getX(), e.getY())) {
                                // pointing the token
                                overToken = ((i + 1) * 10) + (j + 1);
                            }
                        }
                    }
                    // checking if this move would be valid
                    moveCheck(overToken, holdToken);
                    // updating the canvas
                    updateLocation(e);
                }
                
            }
            
        }
        
        // a pressed mouse button is released
        public void mouseReleased (MouseEvent e) {
            
            if (e.getButton() == MouseEvent.BUTTON1) {
                
                // getting the mouse coordinates
                mouseX = e.getX();
                mouseY = e.getY();
                
                // is no popup visible?
                if (showPopup == 0) {

                    // is a game running and is this game not paused and is a token
                    // holded by the mouse?
                    if ((runGame) && (!runPause) && (holdToken != 0)) {
                        // checking if this move would be valid
                        moveCheck(overToken, holdToken);
                        // this move is valid
                        if ((moveValid) && (overToken != holdToken)) {
                            // getting the tokens
                            int ocx = getTokenX(overToken);
                            int ocy = getTokenY(overToken);
                            int hcx = getTokenX(holdToken);
                            int hcy = getTokenY(holdToken);
                            // expanding the undo string
                            strUndo = Integer.toString(holdToken) + Integer.toString(overToken) + Integer.toString(valToken[ocx][ocy]) + strUndo;
                            // set the holded token to the overed place
                            valToken[ocx][ocy] = valToken[hcx][hcy];
                            // clean the place the holded token came from
                            valToken[hcx][hcy] = 0;
                            // count down the number of available tokens
                            countToken--;
                            actButton[2] = 1;
                            // setting the marker for move check
                            moveValid = false;
                            // playing the drop sound
                            doSound(sndDrop);
                        }
                        // this move is not valid
                        else {
                            // setting the marker for move check
                            moveValid = false;
                            // playing the nodrop sound
                            doSound(sndNodrop);
                        }
                    }
                    
                }
                
                // checking all buttons of the visible popup
                for (int i = begPopup[showPopup]; i < (begPopup[showPopup] + btnPopup[showPopup]); i++) {
                    if (rctButton[i].contains(e.getX(), e.getY())) {
                        if ((holdButton == (i + 1)) && (actButton[i] == 3)) {
                            // which action is to be called?
                            int p = 0;
                            if (lenScore > 0) {
                                p = lenScore - pagScore + 1;
                                if (p > lenScore) { p = lenScore; } 
                                if (p < 1) { p = 1; } 
                            }
                            switch (holdButton) {
                            case 1:  
                                // main - new
                                clickButtonNew();
                                break;
                            case 2:  
                                // main - pause
                                doSound(sndClick);
                                clickButtonPause();
                                break;
                            case 3:  
                                // main - undo
                                clickButtonUndo();
                                break;
                            case 4:  
                                // main - highscore
                                doSound(sndClick);
                                clickButtonScore();
                                break;
                            case 5:  
                                // main - sound
                                clickButtonSound();
                                doSound(sndClick);
                                break;
                            case 6:  
                                // main - fullscreen
                                doSound(sndClick);
                                clickButtonFullscreen();
                                break;
                            case 7:  
                                // main - help
                                doSound(sndClick);
                                clickButtonHelp();
                                break;
                            case 8:  
                                // main - about
                                doSound(sndClick);
                                clickButtonInfo();
                                break;
                            case 9:  
                                // main - quit
                                doSound(sndClick);
                                clickButtonQuit();
                                break;
                            case 10: 
                                // confirm quit - no
                                doSound(sndClick);
                                showPopup = 0;
                                break;
                            case 11: 
                                // confirm quit - yes
                                doSound(sndClick);
                                confirmQuit = true;
                                clickButtonQuit();
                                break;
                            case 12: 
                                // gameover - undo
                                showPopup = 0;
                                doPause();
                                clickButtonUndo();
                                break;
                            case 13: 
                                // gameover - finish
                                doSound(sndClick);
                                gameOver();
                                break;
                            case 14: 
                                // input - ok
                                doSound(sndClick);
                                showPopup = 4;
                                break;
                            case 15: 
                                // highscore - delete
                                doSound(sndClick);
                                showPopup = 7;
                                break;
                            case 16: 
                                // highscore - finish
                                doSound(sndClick);
                                mrkScore = 0;
                                showPopup = 0;
                                break;
                            case 17: 
                                // highscore - |<
                                doSound(sndClick);
                                if (lenScore > 0) {
                                    shwScore = 1;
                                }
                                break;
                            case 18: 
                                // highscore - <<
                                doSound(sndClick);
                                if (lenScore > 0) {
                                    shwScore -= 10;
                                    if (shwScore < 1) { shwScore = 1; }
                                }
                                break;
                            case 19: 
                                // highscore - <
                                doSound(sndClick);
                                if (lenScore > 0) {
                                    shwScore--;
                                    if (shwScore < 1) { shwScore = 1; }
                                }
                                break;
                            case 20: 
                                // highscore - >
                                doSound(sndClick);
                                if (lenScore > 0) {
                                    shwScore++;
                                    if (shwScore > p) { shwScore = p; }
                                }
                                break;
                            case 21: 
                                // highscore - >>
                                doSound(sndClick);
                                if (lenScore > 0) {
                                    shwScore += 10;
                                    if (shwScore > p) { shwScore = p; }
                                }
                                break;
                            case 22: 
                                // highscore - >|
                                doSound(sndClick);
                                if (lenScore > 0) {
                                    shwScore = p;
                                }
                                break;
                            case 23: 
                                // fullscreen - ok
                                doSound(sndClick);
                                showPopup = 0;
                                break;
                            case 24: 
                                // confirm cancel game - no
                                doSound(sndClick);
                                if (!tmpPause) { doPause(); }
                                showPopup = 0;
                                break;
                            case 25: 
                                // confirm cancel game - yes
                                doSound(sndClick);
                                showPopup = 0;
                                gameOver();
                                break;
                            case 26: 
                                // confirm highscore reset - no
                                doSound(sndClick);
                                showPopup = 4;
                                break;
                            case 27: 
                                // confirm highscore reset - yes
                                doSound(sndClick);
                                if (lenScore > 0) {
                                    mrkScore = 0;
                                    lenScore = 0;
                                    dataScore.clear();
                                }
                                showPopup = 4;
                                break;
                            case 28: 
                                // about dialog - license
                                doSound(sndClick);
                                try { BrowserLauncher.openURL(getFileURL("license")); } catch (Exception x) {}
                                break;
                            case 29: 
                                // about dialog - close
                                doSound(sndClick);
                                showPopup = 0;
                                break;
                            case 30: 
                                // confirm timeout - finish
                                doSound(sndClick);
                                showPopup = 0;
                                gameOver();
                                break;
                            }
                        }
                    }
                    // is the button enabled?
                    if (actButton[i] > 0) {
                        // setting the state of the button
                        actButton[i] = 1;
                    }
                }
                
            }

            if (e.getButton() == MouseEvent.BUTTON3) {
                
                // was the click within the playing field?
                if (rctField.contains(e.getX(), e.getY())) {
                    // is the undo button enabled?
                    if ((actButton[2] == 3) && (holdButton == 101)) {
                        // doing the undo
                        clickButtonUndo();
                    }
                }
                
            }

            if ((runGame) && (!runPause)) {
                // is there another move available?
                moveValid = false;
                for (int i = 0; i < 5; i++) {
                    for (int j = 0; j < 5; j++) {
                        if (valToken[i][j] != 0) {
                            for (int k = i; k < 5; k++) {
                                for (int l = j; l < 5; l++) {
                                    if (valToken[k][l] != 0) {
                                        if (!moveValid){
                                            int token1 = ((i + 1) * 10) + (j + 1);
                                            int token2 = ((k + 1) * 10) + (l + 1);
                                            if (token1 != token2) {
                                                moveCheck(token1, token2);
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                if (!moveValid) {
                    // no more moves available
                    doPause();
                    showPopup = 2;
                }
            }
            
            // releasing the undo button if it's necessary
            if (actButton[2] == 3) { actButton[2] = 1; }
            // releasing a holded token
            holdToken = 0;
            // resetting the marker for a valid move 
            moveValid = false;
            // releasing a holded button
            holdButton = 0;
            // updating the cursor
            updateCursor(e);
            // updating the canvas
            updateLocation(e);
            
        }
        
        // the mouse is moved while no mouse button is pressed
        public void mouseMoved (MouseEvent e) {
            
            // updating the cursor
            updateCursor(e);
            
        }

        // the mouse is clicked (this is required)
        public void mouseClicked (MouseEvent e) {}
        // the mouse leaves the canvas (this is required)
        public void mouseExited (MouseEvent e) {}
        // the mouse enteres the canvas (this is required)
        public void mouseEntered (MouseEvent e) {}
        
        // checking if a move is possible
        private void moveCheck (int token1, int token2) {
            
            // is no popup visible?
            if (showPopup == 0) {
                
                // is a game running and is the game not paused?
                if ((runGame) && (!runPause)) {
                    moveValid = false;
                    
                    // getting the coordinates of the holded token and the pointed token
                    int ocx = getTokenX(token1);
                    int ocy = getTokenY(token1);
                    int hcx = getTokenX(token2);
                    int hcy = getTokenY(token2);

                    // is the pointed token in the same column or in the same row
                    // as the holded token?
                    if ((ocx == hcx) || (ocy == hcy)) {
                        // getting the values of the holded token and the pointed token
                        int ov = valToken[ocx][ocy];
                        int hv = valToken[hcx][hcy];
                        int ovx = getTokenX(ov);
                        int ovy = getTokenY(ov);
                        int hvx = getTokenX(hv);
                        int hvy = getTokenY(hv);
                        
                        // got the pointed token the same number or the same color
                        // as the holded token?
                        if ((ovx == hvx) || (ovy == hvy)) {
                            moveValid = true;
                        }
                    }
                }
                
            }
            
        }
        
        // updating the cursor = make sure the right cursor is shown
        private void updateCursor (MouseEvent e) {
            
            // the cursor type
            // 0 = the arrow cursor, default
            // 1 = the hand cursor (mouse is pointing a button or a token)
            // 2 = the move cursor (mouse is holding a token)
            int mc = 0;
            
            // the button which the mouse is pointing
            // 0 = no button
            // 1 = new game
            // 2 = undo
            // 3 = highscores
            // 4 = language
            // 5 = sound
            // 6 = help
            // 7 = about
            // 8 = quit
            int ob = 0;
            
            // checking all buttons of the visible popup
            for (int i = begPopup[showPopup]; i < (begPopup[showPopup] + btnPopup[showPopup]); i++) {
                if (rctButton[i].contains(e.getX(), e.getY())) {
                    // is the button enabled and is it not holded and is no button holded?
                    if ((actButton[i] > 0) && (actButton[i] < 3) && (holdButton == 0)) {
                        // the hand cursor is needed
                        mc = 1;
                        // what button is pointed?
                        ob = i + 1;
                        // setting the state of the button
                        actButton[i] = 2;
                    }
                }
                else {
                    // is the button enabled and is it not holded?
                    if ((actButton[i] > 0) && (actButton[i] < 3)) {
                        // setting the state of the button
                        actButton[i] = 1;
                    }
                }
            }
            
            // is a button pointed?
            if (ob != overButton) {
                // what button is pointed?
                overButton = ob;
                // update the canvas
                updateLocation(e);
            }
            
            // is no popup visible?
            if (showPopup == 0) {
                
                // is a game running and is the game not paused?
                if ((runGame) && (!runPause)) {
                    // is no token holded by the mouse?
                    if (holdToken == 0) {
                        // checking all tokens
                        for (int i = 0; i < 5; i++) {
                            for (int j = 0; j < 5; j++) {
                                if (rctToken[i][j].contains(e.getX(), e.getY())) {
                                    if (valToken[i][j] != 0) {
                                        // the hand cursor is needed
                                        mc = 1;
                                    }
                                }
                            }
                        }
                    }
                    // is a token holded by the mouse?
                    else {
                        // the move cursor is needed
                        mc = 2;
                    }
                }
                
            }
            
            // setting the cursor which is needed
            switch (mc) {
            case 0: setCursor(curDefault); break;
            case 1: setCursor(curHand); break;
            case 2: setCursor(curMove); break;
            }
            
        }
    
        // update the canvas
        private void updateLocation (MouseEvent e) {
            
            // repaint the canvas
            repaint();
            
        }
    
        // paint the canvas
        public void paint (Graphics g) {
            
            // do the canvas update
            update(g);
            
        }

        // make the canvas update
        public void update (Graphics g) {
            
            // the canvas' graphics object
            Graphics2D g2 = (Graphics2D)g;
            // make sure quality alpha
            big.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION, RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
            // make sure text antialiasing
            big.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
            
            // drawing the background
            big.drawImage(imgBackground, 0, 0, this);
            
            // drawing the buttons
            for (int i = begPopup[0]; i < (begPopup[0] + btnPopup[0]); i++) { writeButton(i); }
            
            // is a game running?
            if (runGame) {
                // draw the label for the token counter
                writeVertical(messages.getString("StatusTokens"), -1, countTokenY + countDigitH + 4);
                // draw the token counter
                int i = countToken / 10; if (i == 0) { i = 10; }
                big.drawImage(imgDigit[i], countTokenX, countTokenY + countDigitH, this);
                big.drawImage(imgDigit[countToken % 10], countTokenX, countTokenY, this);
                // draw the label for the time counter
                writeVertical(messages.getString("StatusTime"), -2, countTimeY + (countDigitH * 4) + 4);
                // draw the time counter
                i = countTimeM / 10; if (i == 0) { i = 10; }
                big.drawImage(imgDigit[i], countTimeX, countTimeY + (countDigitH * 4), this);
                big.drawImage(imgDigit[countTimeM % 10], countTimeX, countTimeY + (countDigitH * 3), this);
                big.drawImage(imgDigit[11], countTimeX, countTimeY + (countDigitH * 2), this);
                big.drawImage(imgDigit[countTimeS / 10], countTimeX, countTimeY + countDigitH, this);
                big.drawImage(imgDigit[countTimeS % 10], countTimeX, countTimeY, this);
                if (runPause) {
                    // draw the pause text
                    writeText (messages.getString("StatusPause"), (int)rctField.getX(), (int)rctField.getY(), (int)rctField.getWidth(), (int)rctField.getHeight(), 0, false, false, fntText);
                }
            }
            else {
                writeText (messages.getString("StatusNogame"), (int)rctField.getX(), (int)rctField.getY(), (int)rctField.getWidth(), (int)rctField.getHeight(), 0, false, false, fntText);
            }
            
            // is a game running?
            if ((runGame) && (!runPause)) {
                for (int i = 0; i < 5; i++) {
                    for (int j = 0; j < 5; j++) {
                        // is this token not holded by the mouse?
                        if (holdToken != ((i + 1) * 10) + (j + 1)) {
                            // drawing the tokens
                            if (valToken[i][j] > 0) {
                                int drawToken = valToken[i][j];
                                if ((overToken == ((i + 1) * 10) + (j + 1)) && (moveValid)) {
                                    big.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.6f));
                                    big.drawImage(imgToken[getTokenX(drawToken)][getTokenY(drawToken)], tokenX[i], tokenY[j], this);
                                    big.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 1.0f));
                                }
                                else {
                                    big.drawImage(imgToken[getTokenX(drawToken)][getTokenY(drawToken)], tokenX[i], tokenY[j], this);
                                }
                            }
                        }
                    }
                }
            }
            
            // is a game running and is the game not paused?
            if ((runGame) && (!runPause)) {
                // is no popup visible and is a token holded by the mouse?
                if ((showPopup == 0) && (holdToken > 0)) {
                    // drawing the holded token at the mouse position
                    int drawToken = valToken[getTokenX(holdToken)][getTokenY(holdToken)];
                    // is the current move valid = draw the token
                    if (moveValid) {
                        big.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.8f));
                        big.drawImage(imgToken[getTokenX(drawToken)][getTokenY(drawToken)], mouseX - (tokenW / 2), mouseY - (tokenW / 2), this);
                        big.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 1.0f));
                    }
                    // is the current move not valid = draw the nodrop image
                    else {
                        big.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.4f));
                        big.drawImage(imgToken[getTokenX(drawToken)][getTokenY(drawToken)], mouseX - (tokenW / 2), mouseY - (tokenW / 2), this);
                        big.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 1.0f));
                        big.drawImage(imgNodrop, mouseX - (nodropW / 2), mouseY - (nodropH / 2), this);
                    }
                }
            }

            // showing the confirm quit dialog
            if (showPopup == 1) {
                
                // making all the background grayscale
                makeGray();
                // the dialog background
                big.drawImage(imgDialog2, questionX, questionY, this);
                // drawing the buttons
                for (int i = begPopup[showPopup]; i < (begPopup[showPopup] + btnPopup[showPopup]); i++) { writeButton(i); }
                // output the message
                writeText (messages.getString("TextQuit"), textDialogX, textDialogY, textDialogW, textDialogH, 0, false, false, fntText);
                
            }
            
            // showing the gameover question dialog
            if (showPopup == 2) {
                
                // making all the background grayscale
                makeGray();
                // the dialog background
                big.drawImage(imgDialog2, questionX, questionY, this);
                // drawing the buttons
                for (int i = begPopup[showPopup]; i < (begPopup[showPopup] + btnPopup[showPopup]); i++) { writeButton(i); }
                // output the message
                writeText (messages.getString("TextGameover1"), textDialogX, textDialogY, textDialogW, textDialogH, -1, false, false, fntText);
                writeText (messages.getString("TextGameover2"), textDialogX, textDialogY, textDialogW, textDialogH, -2, false, false, fntText);
                
            }
            
            // showing the input name dialog
            if (showPopup == 3) {
                
                // making all the background grayscale
                makeGray();
                // the dialog background
                big.drawImage(imgDialog1, questionX, questionY, this);
                // drawing the buttons
                for (int i = begPopup[showPopup]; i < (begPopup[showPopup] + btnPopup[showPopup]); i++) { writeButton(i); }
                // output the message
                writeText (messages.getString("TextName"), textDialogX, textDialogY, textDialogW, textDialogH, -1, false, false, fntText);
                writeText (strName, textDialogX, textDialogY, textDialogW, textDialogH, -2, true, false, fntText);
                
            }
            
            // showing the gameover question dialog
            if ((showPopup == 4) || (showPopup == 7)) {
                
                // making all the background grayscale
                if (showPopup == 4) { makeGray(); }
                // the dialog background
                big.drawImage(imgHighscore, highscoreX, highscoreY, this);
                // drawing the buttons
                for (int i = begPopup[4]; i < (begPopup[4] + btnPopup[4]); i++) { writeButton(i); }
                // output the highscorelist
                writeText (messages.getString("HeadHighscoreName"), textHighNameX, textHighHeadY, textHighNameW, textHighHeadH, 0, false, false, fntText);
                writeText (messages.getString("HeadHighscoreTokens"), textHighTokensX, textHighHeadY, textHighTokensW, textHighHeadH, 0, false, false, fntText);
                writeText (messages.getString("HeadHighscoreTime"), textHighTimeX, textHighHeadY, textHighTimeW, textHighHeadH, 0, false, false, fntText);
                String[] setScore = new String[3];
                int j = 0;
                int k = shwScore + pagScore - 1;
                if (k > lenScore) { k = lenScore; }
                for (int i = shwScore; i <= k; i++) {
                    j++;
                    boolean m = (i == mrkScore);
                    setScore = (String[])dataScore.get(i - 1);
                    writeText (Integer.toString(i), textHighNumX, textHighLineY, textHighNumW, textHighLineH, j, false, m, fntText);
                    writeText (setScore[0], textHighNameX, textHighLineY, textHighNameW, textHighLineH, j, false, m, fntText);
                    writeText (setScore[1], textHighTokensX, textHighLineY, textHighTokensW, textHighLineH, j, false, m, fntText);
                    writeText (setScore[2], textHighTimeX, textHighLineY, textHighTimeW, textHighLineH, j, false, m, fntText);
                }
                int a = shwScore;
                if (lenScore == 0) { a = 0; }
                int e = shwScore + 9;
                if (e > lenScore) { e = lenScore; }
                writeText (Integer.toString(a) + "-" + Integer.toString(e) + " / " + Integer.toString(lenScore), textHighFootX, textHighFootY, textHighFootW, textHighFootH, 0, false, false, fntText);
                
            }
            
            // is the fullscreen error message visible?
            if (showPopup == 5) {
                
                // making all the background grayscale
                makeGray();
                // the dialog background
                big.drawImage(imgDialog1
                        , questionX, questionY, this);
                // drawing the buttons
                for (int i = begPopup[showPopup]; i < (begPopup[showPopup] + btnPopup[showPopup]); i++) { writeButton(i); }
                // output the message
                writeText (messages.getString("TextFullscreen1"), textDialogX, textDialogY, textDialogW, textDialogH, -1, false, false, fntText);
                writeText (messages.getString("TextFullscreen2"), textDialogX, textDialogY, textDialogW, textDialogH, -2, false, false, fntText);
                
            }
            
            // showing the confirm quit dialog
            if (showPopup == 6) {
                
                // making all the background grayscale
                makeGray();
                // the dialog background
                big.drawImage(imgDialog2, questionX, questionY, this);
                // drawing the buttons
                for (int i = begPopup[showPopup]; i < (begPopup[showPopup] + btnPopup[showPopup]); i++) { writeButton(i); }
                // output the message
                writeText (messages.getString("TextCancel1"), textDialogX, textDialogY, textDialogW, textDialogH, -1, false, false, fntText);
                writeText (messages.getString("TextCancel2"), textDialogX, textDialogY, textDialogW, textDialogH, -2, false, false, fntText);
                
            }
            
            // showing the confirm highscore reset dialog
            if (showPopup == 7) {
                
                // making all the background grayscale
                makeGray();
                // the dialog background
                big.drawImage(imgDialog2, questionX, questionY, this);
                // drawing the buttons
                for (int i = begPopup[showPopup]; i < (begPopup[showPopup] + btnPopup[showPopup]); i++) { writeButton(i); }
                // output the message
                writeText (messages.getString("TextHighReset1"), textDialogX, textDialogY, textDialogW, textDialogH, -1, false, false, fntText);
                writeText (messages.getString("TextHighReset2"), textDialogX, textDialogY, textDialogW, textDialogH, -2, false, false, fntText);
                
            }
            
            // showing the about dialog
            if (showPopup == 8) {
                
                // making all the background grayscale
                makeGray();
                // the dialog background
                big.drawImage(imgAbout, aboutX, aboutY, this);
                // drawing the buttons
                for (int i = begPopup[showPopup]; i < (begPopup[showPopup] + btnPopup[showPopup]); i++) { writeButton(i); }
                // drawing the applications icon
                big.drawImage(imgBigIcon, aboutX + 50, aboutY + 50, this);
                // output the application title and the copyright
                writeText (strAppTitle + " " + strAppVersion, aboutX + 100, aboutY + 47, 290, 22, 0, false, true, fntTitle);
                writeText ("(C) 2001-05 by Frank Hollwitz", aboutX + 100, aboutY + 69, 290, 15, 0, false, true, fntCopyr);
                for (int i = 1; i <= 6; i++) {
                      writeText (messages.getString("TextAbout" + Integer.toString(i)), aboutX + 50, aboutY + 100, 340, textHighLineH, i, false, false, fntText);
                }
                
            }
            
            // showing the confirm quit dialog
            if (showPopup == 9) {
                
                // making all the background grayscale
                makeGray();
                // the dialog background
                big.drawImage(imgDialog1, questionX, questionY, this);
                // drawing the buttons
                for (int i = begPopup[showPopup]; i < (begPopup[showPopup] + btnPopup[showPopup]); i++) { writeButton(i); }
                // output the message
                writeText (messages.getString("TextTimeout1"), textDialogX, textDialogY, textDialogW, textDialogH, -1, false, false, fntText);
                writeText (messages.getString("TextTimeout2"), textDialogX, textDialogY, textDialogW, textDialogH, -2, false, false, fntText);
                
            }
            
            // copiing the double buffered image to the canvas
            g2.drawImage(bi, 0, 0, this);
            g2.dispose();
            
        }
        
        // making all the background grayscale
        private void makeGray () {
            
            BufferedImage imgBW = new BufferedImage(bi.getWidth(), bi.getHeight(), BufferedImage.TYPE_BYTE_GRAY);
            imgBW.getGraphics().drawImage(bi, 0, 0, null);
            big.drawImage(imgBW, 0, 0, this);
            
        }
        
        // writing the texts
        private void writeText (String s, int x, int y, int w, int h, int l, boolean b, boolean m, Font f) {

            // setting the font and calculating the font metrics
            big.setFont(f);
            FontMetrics metText = big.getFontMetrics();
            int fh = metText.getHeight();
            int fa = metText.getAscent();
            int fw = metText.stringWidth(s);
            // which line of how many lines is this?
            int line = 0;
            if (l == -2) { line = fh / 2; }
            if (l == -1) { line = 0 - (fh / 2); }
            if (l == 0) { line = 0; }
            if (l > 0) { line = (l - 1) * (fh + 2); }
            // is this line an input line? 
            if (b) {
                // this is an input line
                big.setColor(colInputShadow);
                big.drawString(s, getTextX(x, w, fw) + 1, getTextY(y, h, fh, fa) + 1 + line);
                big.setColor(colInput);
                big.drawString(s, getTextX(x, w, fw), getTextY(y, h, fh, fa) + line);
                big.setColor(colTextShadow);
                big.drawString("|", getTextX(x, w, fw) + 1 + fw, getTextY(y, h, fh, fa) + 1 + line);
                big.setColor(colText);
                big.drawString("|", getTextX(x, w, fw) + fw, getTextY(y, h, fh, fa) + line);
            }
            else {
                if (m) {
                    // this is a marked highscore line
                    big.setColor(colInputShadow);
                    big.drawString(s, getTextX(x, w, fw) + 1, getTextY(y, h, fh, fa) + 1 + line);
                    big.setColor(colInput);
                    big.drawString(s, getTextX(x, w, fw), getTextY(y, h, fh, fa) + line);
                }
                else {
                    // this is no input line and no marked highscore
                    big.setColor(colTextShadow);
                    big.drawString(s, getTextX(x, w, fw) + 1, getTextY(y, h, fh, fa) + 1 + line);
                    big.setColor(colText);
                    big.drawString(s, getTextX(x, w, fw), getTextY(y, h, fh, fa) + line);
                }
            }
            
        }
        
        // writing the vertical texts
        private void writeVertical (String s, int x, int y) {
            
            // setting the font and calculating the font metrics
            big.setFont(fntText);
            FontMetrics metText = big.getFontMetrics();
            int fh = metText.getHeight();
            int fa = metText.getAscent();
            int fw = metText.stringWidth(s);
            // needs the text to be centered?
            if (x == 0) { x = ((textStatusW - fh) / 2) + textStatusX; }
            if (x == -1) { x = ((textStatusW - (fh * 2)) / 3) + textStatusX; }
            if (x == -2) { x = ((textStatusW - (fh * 2)) / 3 * 2) + fh + textStatusX; }
            if (y == 0) { y = ((mainH - fw) / 2) - fh; }
            // draw the text
            AffineTransform rotAT = new AffineTransform(saveAT);
            rotAT.rotate(Math.toRadians(270), x + fa, y + fw);
            big.setTransform(rotAT);
            big.setColor(colTextShadow);
            big.drawString(s, x - 1, y + fw + 1);
            big.setColor(colText);
            big.drawString(s, x, y + fw);
            big.setTransform(saveAT);
            
        }
        
        // drawing the buttons
        private void writeButton (int i) {
            
            // get width and height of the buttons
            int w = buttonW;
            if ((i >= 16) && (i < 22)) { w = buttonWs; }
            int h = buttonH;
            // enable switching the first button between newgame/pause/resume 
            String txtButton = messages.getString(resButton[i]);
            if (i == 0) { if (runGame) { txtButton = messages.getString("ButtonMainStopgame"); } else { txtButton = messages.getString("ButtonMainNewgame"); } }
            if (i == 1) { if (runPause) { txtButton = messages.getString("ButtonMainResume"); } else { txtButton = messages.getString("ButtonMainPause"); } }
            if (i == 4) { if (playSound) { txtButton = "[X] " + txtButton; } else { txtButton = "[_] " + txtButton; } }
            if (i == 5) { if (isFullscreen) { txtButton = "[X] " + txtButton; } else { txtButton = "[_] " + txtButton; } }
            // setting the font and calculating the font metrics
            big.setFont(fntButton);
            FontMetrics metButton = big.getFontMetrics();
            int fh = metButton.getHeight();
            int fa = metButton.getAscent();
            int fw = metButton.stringWidth(txtButton);
            // drawing the light buttons text
            big.setColor(colButtonLight);
            big.drawString(txtButton, getTextX(buttonX[i], w, fw) - 1, getTextY(buttonY[i], h, fh, fa) - 1);
            // drawing the buttons shadow text
            big.setColor(colButtonShadow);
            big.drawString(txtButton, getTextX(buttonX[i], w, fw) + 1, getTextY(buttonY[i], h, fh, fa) + 1);
            // drawing the buttons text
            switch (actButton[i]) {
            case 0: big.setColor(colButtonDisabled); break;
            case 1: big.setColor(colButtonNormal); break;
            case 2: big.setColor(colButtonOver); break;
            case 3: big.setColor(colButtonClick); break;
            }
            big.drawString(txtButton, getTextX(buttonX[i], w, fw), getTextY(buttonY[i], h, fh, fa));
            
        }
        
        // calculating the x-coordinate for text outputs
        private int getTextX (int x, int w, int t) {
            return (x + ((w - t) / 2));
        }
        
        // calculating the y-coordinate for text outputs
        private int getTextY (int y, int h, int t, int a) {
            return (y + a + ((h - t) / 2));
        }
        
    }
    
    // the timer thread = the stopwatch
    private class boxit_timer extends Thread {
        
        public void run() {
            
            while ( !isInterrupted() ) {
                
                try { Thread.sleep(1000); } catch (InterruptedException e) { interrupt(); }
                
                if ((runGame) && (!runPause)) {
                    
                    countTimeS++;
                    if (countTimeS == 60) {
                        countTimeM++;
                        countTimeS = 0;
                    }
                    
                    if ((countTimeM == 60) && (countTimeS >= 0)) { timeout(); }
                
                    canMain.repaint();
                }
                
                this.run();
                
            }
            
        }
        
    }
    
}

